# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain
   :platform: Unix, Windows
   :synopsis: Canonical Text Service and Canonical URN tool suite for Python

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

__name__ = "MyCapytain"
__version__ = "0.1.3"
__all__ = ["common", "endpoints", "resources"]
