# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.endpoints.ahab
   :synopsis: Ahab endpoint implementation

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
import MyCapytain.endpoints.proto
import requests


class Ahab(MyCapytain.endpoints.proto.Ahab):

    """ 
    Basic integration of the proto.CTS abstraction
    """
    pass
